#13/01/2025

#other useful functions to use with the EOS and EOS data
import numpy as np
import pandas as pd
from scipy.spatial import distance
import warnings
from scipy import optimize

from .symbols import *

#find multiple roots of a function between a specific bracket (univariate function only)
def multi_root(f: callable = None, bracket = None, args: tuple = (), n: int = 1000) -> np.ndarray:
    """ Find all roots of f in `bracket`, given that resolution `n` covers the sign change.
    Fine-grained root finding is performed with `scipy.optimize.root_scalar`.
    Parameters
    ----------
    f: Callable
        Function to be evaluated
    bracket: Sequence of two floats
        Specifies interval within which roots are searched in log space.
    args: Iterable, optional
        Iterable passed to `f` for evaluation
    n: int
        Number of points sampled equidistantly from bracket to evaluate `f`.
        Resolution has to be high enough to cover sign changes of all roots but not finer than that.
        Actual roots are found using `scipy.optimize.root_scalar`.
    Returns
    -------
    roots: np.ndarray
        Array containing all unique roots that were found in `bracket`.
    """
    # Evaluate function in given bracket

    # x = np.linspace(*bracket, int(n))
    x = np.logspace(*bracket, int(n))

    y = f(x, *args)

    # Find where adjacent signs are not equal
    sign_changes = np.where(np.sign(y[:-1]) != np.sign(y[1:]))[0]

    # Find roots around sign changes
    root_finders = (
        optimize.root_scalar(
            method="bisect",
            f=f,
            args=args,
            bracket=(x[s], x[s+1]),
            xtol=1e-12, rtol=1e-12
        )
        for s in sign_changes
    )

    roots = np.array([
        r.root if r.converged else np.nan
        for r in root_finders
    ])


    if np.any(np.isnan(roots)):
        warnings.warn("Not all root finders converged for estimated brackets! Maybe increase resolution `n`.")
        roots = roots[~np.isnan(roots)]

    roots_unique = np.unique(roots)
    if len(roots_unique) != len(roots):
        warnings.warn("One root was found multiple times. "
                    "Try to increase or decrease resolution `n` to see if this warning disappears.")

    #check if the roots correspond to actual roots or a discontinuity
    #if inputting the root into the function results in a distance higher than 1 then assume its not a root and assign None
    roots_unique[abs(f(roots_unique, *args)) > 1e0] = None
    roots_unique = roots_unique[~np.isnan(roots_unique)]

    # sort the roots in ascending order
    roots_unique = np.sort(roots_unique)


    return roots_unique   

#choose the best root from a list by knowing the experimental value
def choose_root(roots: list, experimental_value=0):
    roots = np.array(roots)

    difference = np.abs(experimental_value-roots)

    try:
        best_index = np.nanargmin(difference)
        best_root = roots[best_index]

    except ValueError:
        best_root = np.nan

    return best_root

#see zandbox for trial and error
def calc_rho_inverse_distance(df_missing_rho, df_experimental, pascals=1e4, neighbours=3):
    #restrict experimental dataframe to Paper, P, T, rho and deep copy it
    df_experimental = df_experimental.copy(deep=True)[["Paper", "P", "T", "rho"]]

    #restrict the other dataframes to Paper, P,T and deep copy it
    df_missing_rho = df_missing_rho.copy(deep=True)[["Paper", "P", "T"]]

    #create og index columns
    df_missing_rho["index"] = df_missing_rho.index
    df_experimental["index"] = df_experimental.index
    
    #merge the two dataframes on pressure and temperature
    df = pd.merge(df_experimental, df_missing_rho, on=["Paper","P", "T"], how="outer")

    #sort by temperature and then by pressure
    df = df.sort_values(by=["T","P"])
    #delete pressure and temperature duplicates
    # df = df.drop_duplicates(subset=["P", "T"])

    #reset the indices
    df = df.reset_index()

    #create existance column
    df["exist"] = df["rho"].notna()

    #reduce the pressure by the pascals
    df["P"] = df["P"]/pascals

    #calculate distances between points
    distances = distance.cdist(df[["T","P"]], df[["T","P"]], "euclidean")

    #multiply the distances by the truth vector
    distances_true = (distances.T * np.array(df["exist"])).T

    #make the distances a dataframe
    df_distances = pd.DataFrame(data=distances_true)

    #get the point numbers for which you would like to linearly interpolate
    point_numbers = df.index[df["exist"]==False]

    #list to keep track of interpolation values
    rho_interp = []

    #for unknown point
    for i in point_numbers:
        distance_column = df_distances[i]

        #get the smallest neighbours nonzero values
        smallest_n = distance_column.loc[distance_column!=0].nsmallest(neighbours)

        #calculate the inverse distance density and add it to the list
        rho_interp +=[sum(df.loc[smallest_n.index, "rho"]*1/smallest_n.values)/sum(1/smallest_n.values)]

    #add the interpolated values to the dataframe
    df.loc[point_numbers,["rho"]] = rho_interp

    #return the pressure back to normal
    df["P"] = df["P"]*pascals

    #drop all the rows where the index_y is Nan
    df = df[df["index_y"].notna()]

    # set index to the index of the dataframe which is missing density
    df = df.set_index("index_y").sort_index().reset_index(drop=True)

    return df[["Paper", "P", "T", "rho"]]   #return the  dataframe but only Paper, P, T, rho

def calc_Psat(fluid, EOS, temperature_list: list):
    """
    Calculate the saturation pressure of a fluid at given temperatures.

    Parameters:
    fluid: The fluid object containing properties like critical and triple point temperatures.
    EOS: The equation of state object used for pressure and fugacity calculations.
    temperature_list (list): A list of temperatures at which to calculate saturation pressures.

    Returns:
    list: A list of calculated saturation pressures corresponding to the input temperatures.
    """
    
    pressure_equation = sp.utilities.lambdify((rho, T, P), P - EOS.pressure)
    fugacity_coefficient_function = sp.utilities.lambdify((rho, T), EOS.fugacity_coefficient)

    def get_pressure_guess(temperature):
        """
        Get an initial pressure guess for a specified temperature.

        Parameters:
        temperature: The temperature for which to estimate the pressure.

        Returns:
        float: An initial pressure guess.
        """
        pressure_guess = fluid.P_t# + (fluid.P_c - fluid.P_t) * (temperature - fluid.T_t) / (fluid.T_c - fluid.T_t)
        n_roots = 0
        while n_roots < 3:
            pressure_guess += 1e2
            density_roots = multi_root(pressure_equation, [1e-4, 1e4], args=(temperature, pressure_guess))
            n_roots = len(density_roots)
        return pressure_guess

    def calc_saturation_pressure(pressure, temperature):
        """
        Calculate the saturation pressure from temperature and initial pressure guess.

        Parameters:
        pressure: The initial pressure guess.
        temperature: The temperature for which to calculate saturation pressure.

        Returns:
        float: The result of the fugacity ratio equation.
        """
        density_roots = multi_root(pressure_equation, [1e-4, 1e4], args=(temperature, pressure))
        fugacities = fugacity_coefficient_function(density_roots, temperature)
        fugacities = fugacities[fugacities != max(fugacities)]
        eq = fugacities[0] / fugacities[1] - 1
        return eq

    P_sat = []
    pressure_guess = get_pressure_guess(temperature_list[0])

    for temperature in temperature_list:
        if temperature > fluid.T_c or temperature < fluid.T_t:
            print("Temperature is outside the critical and triple point bounds")
            break
        try:
            sol = optimize.root(calc_saturation_pressure, x0=[pressure_guess], args=(temperature,), method='broyden1', options={"xtol": 1e2})
            solution = sol.x
            pressure_guess = sol.x[0]
        except IndexError:
            print("IndexError: no solution has been found due to the root finding method reaching a point with less than two density roots")
            print("Assigning 0 to return value")
            solution = [0]
        P_sat += list(solution)

    return P_sat





__all__ = ["PureFluid",
           "symbols",
           "storeroom"]

import importlib

from .purefluid import PureFluid
from .symbols import *

# Lazy imports
def __getattr__(name):
    # 2. Check if the user is asking for 'storeroom'
    if name == "storeroom":
        # 3. Only NOW do we import the subpackage
        # print("Lazy loading 'storeroom' subpackage...") # Debug print
        return importlib.import_module(f".{name}", __name__)
    
    # 4. Standard error if they ask for something that doesn't exist
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")





# 30/09/2025

from .symbols import *

# non dimensional Helmholtz energy for SAFT EOS
class AlphaRSAFT():
    
    def __init__(self, epsilon:float, sigma:float, m:float, epsilon_AB:float, k_AB:float, M:int, x_p:float, assocation_scheme:str,
                 alpha_hs, alpha_chain, alpha_disp, alpha_assoc, alpha_multipolar):

        self.epsilon = epsilon

        self.sigma = sigma
        
        self.m = m

        self.epsilon_AB = epsilon_AB

        self.k_AB = k_AB

        self.M = M

        self.x_p = x_p

        self.assocation_scheme = assocation_scheme

        self.alpha_hs = alpha_hs
        self.alpha_chain = alpha_chain
        self.alpha_disp = alpha_disp
        self.alpha_assoc = alpha_assoc
        self.alpha_multipolar = alpha_multipolar

        self.alpha_r = alpha_hs + alpha_chain + alpha_disp + alpha_assoc + alpha_multipolar

        

    # for a pure fluid (fluid is just a placeholder argument)
    @classmethod
    def for_purefluid(cls, fluid, 
                      epsilon:float = 0, sigma:float = 0, m:float = 0, 
                      epsilon_AB:float = 0, k_AB:float = 0, M:int = 0, 
                      x_p:float = 0, 
                      a:list=[], b:list=[], 
                      association_scheme:str = "", 
                      alpha_hs = None, alpha_chain=None, alpha_disp = None, alpha_assoc=None, alpha_multipolar=None,
                      **kwargs):
        
        # values used in most terms
        d = sigma * (1 - 0.12 * sp.exp(-3*epsilon/T))

        zeta_n = [pi/6*rho*m*d**n for n in range(0, 4)]

        eta = zeta_n[3]

        # print(epsilon, sigma, m, epsilon_AB, k_AB, M, a, b)
        # hard sphere
        if alpha_hs == None:
            alpha_hs = cls.alpha_hs(m, zeta_n).subs([(rho, rho*N_av*1e-30)])

        # chain
        if alpha_chain ==None:
            alpha_chain = cls.alpha_chain(m, d, zeta_n).subs([(rho, rho*N_av*1e-30)])

        # dispersion
        if alpha_disp == None:
            m2e1sigma3 = m**2*epsilon/T*sigma**3
            m2e2sigma3 = m**2*(epsilon/T)**2*sigma**3

            alpha_disp = cls.alpha_disp(m, eta, a, b, m2e1sigma3, m2e2sigma3).subs([(rho, rho*N_av*1e-30)])

        # association
        if alpha_assoc == None:
            g_hs = 1/(1-zeta_n[3]) + d/2* 3*zeta_n[2]/(1-zeta_n[3])**2 + d**2/4 * 2*zeta_n[2]**2/(1-zeta_n[3])**3

            Delta = g_hs*(sp.exp(epsilon_AB/T)-1) * (sigma**3 * k_AB)
            
            alpha_assoc = cls.alpha_assoc(association_scheme, Delta).subs([(rho, rho*N_av*1e-30)])

        # multipolar
        if alpha_multipolar ==None:
            alpha_multipolar = cls.alpha_multipolar(fluid, sigma, x_p).subs([(rho, rho*N_av*1e-30)])

        return cls(epsilon, sigma, m, epsilon_AB, k_AB, M, x_p, association_scheme, alpha_hs, alpha_chain, alpha_disp, alpha_assoc, alpha_multipolar)

   
    ################################################# Static methods
    # hard sphere
    @staticmethod
    def alpha_hs(m, zeta_n):
        # from [0329]
        alpha_hs = m / zeta_n[0] * ( 3*zeta_n[1]*zeta_n[2]/(1-zeta_n[3]) + zeta_n[2]**3/(zeta_n[3]*(1-zeta_n[3])**2) + (zeta_n[2]**3/zeta_n[3]**2 -zeta_n[0])*sp.log(1-zeta_n[3]))
        # print(f"alpha_hs = {alpha_hs}")
        return alpha_hs
    
    # chain
    @staticmethod
    def alpha_chain(m, d, zeta_n):
        g_hs = 1/(1-zeta_n[3]) + d/2* 3*zeta_n[2]/(1-zeta_n[3])**2 + d**2/4 * 2*zeta_n[2]**2/(1-zeta_n[3])**3

        alpha_chain = -(m-1)*sp.log(g_hs)
        # print(f"alpha_chain = {alpha_chain}")
        return alpha_chain
    
    # dispersion
    @staticmethod
    def alpha_disp(m, eta, a:list, b:list, m2e1sigma3, m2e2sigma3):

        C1 = (1 + m*(8*eta - 2*eta**2)/(1-eta)**4 + (1-m)* (20*eta-27*eta**2 + 12*eta**3 -2*eta**4)/((1-eta)*(2-eta))**2)**(-1)
        # print(f"C1 = {C1}")

        a_i = [a[i][0] + (m-1)/m*a[i][1] + (m-1)/m * (m-2)/m * a[i][2] for i in range(0,7)]

        I1 = sum([a_i[i] *eta**i for i in range (0, 7)])
        # print(f"I1 = {I1}")

        b_i = [b[i][0] + (m-1)/m*b[i][1] + (m-1)/m * (m-2)/m * b[i][2] for i in range(0,7)]

        I2 = sum([b_i[i] *eta**i for i in range (0, 7)])
        # print(f"I2 = {I2}")

        alpha_disp = -2*pi*rho*I1* m2e1sigma3 - pi*rho*m*C1*I2* m2e2sigma3
        # print(f"alpha_disp = {alpha_disp}")
        return alpha_disp
    
    # association
    @staticmethod
    def alpha_assoc(association_scheme: str ="", Delta: sp.core.add.Add = 0):
        ########################################### define association scheme methods
        def assoc_2B(Delta):
            # source = [0328]
            X_A = (-1 + (1+4*rho*Delta)**0.5)/(2*rho*Delta)
            X_B = X_A
            return [X_A, X_B]

        def assoc_4B(Delta):
            # source = [0328]
            X_A = (-(1 - 2 * rho * Delta) + ((1 + 2 * rho * Delta)**2 + 4 * rho *Delta)**0.5)/(6 * rho * Delta)
            X_B = X_A
            X_C = X_A
            X_D = 3*X_A -2
            return [X_A, X_B, X_C, X_D]            

        def assoc_3B(Delta):
            # source = [0328]
            X_A = (-(1 - rho * Delta) + ((1 + rho * Delta)**2 + 4 * rho *Delta)**0.5)/(4 * rho * Delta)
            X_B = X_A
            X_C = 2 * X_A -1
            return [X_A, X_B, X_C]            

        def assoc_4C(Delta):
            # source = [0328]
            X_A = (-1 + (1 + 8 * rho * Delta)**0.5)/(4 * rho * Delta)
            X_B = X_A
            X_C = X_A
            X_D = X_A
            return [X_A, X_B, X_C, X_D]
        
       
            
        ########################################## select association scheme
        if association_scheme=="2B":
            scheme = assoc_2B
            print("association scheme 2B was selected")

        elif association_scheme=="4B":
            scheme = assoc_4B
            print("association scheme 4B was selected")

        elif association_scheme=="3B":
            scheme = assoc_3B 
            print("association scheme 3B was selected")

        elif association_scheme=="4C":
            scheme = assoc_4C
            print("association scheme 4C was selected")

        elif association_scheme=="":   # for if the molecule does not associate, like hydrogen
            print("no association scheme was selected")
            return sp.simplify(0)

        else:
            raise(ValueError("Please select a valid association scheme..."))

        ########################################## calculate the association alpha
           
        X = scheme(Delta)
        
        alpha_assoc = sum([sp.log(X_A) - X_A/2 + 0.5 for X_A in X])
        # print(f"alpha_assoc = {alpha_assoc}")
        return alpha_assoc
    
    # multipolar
    @staticmethod
    def alpha_multipolar(fluid, sigma, x_p):
        mu = fluid.dipole   # in Debyes

        # non dimensional variables
        rho_star = rho*sigma**3                                         # non dimensional number density

        mu_star = mu*1e-18/(k_b_Gaussian*T*(sigma*1e-8)**3)**0.5
        T_star = 1/mu_star**2

        #I2 and I3 expressions which simplify the calculation of the multipolar term, these are different from the dispersion ones
        I2 = (1 - 0.3618*rho_star - 0.3205*rho_star**2 + 0.1078*rho_star**3)/(1-0.5236*rho_star)**2

        I3 = (1 + 0.62378*rho_star - 0.11658*rho_star**2)/(1-0.59056*rho_star + 0.20059*rho_star**2)

        #now calculate the multipolar term
        alpha_multipolar = (-2*pi/9 * (1/T_star)**2 * rho_star * x_p**2 * I2) / (1 + 5*pi/36 * 1/T_star * rho_star * x_p *I3/I2)

        return alpha_multipolar
    
    


"""Lazy-load storeroom submodules/objects on attribute access (PEP 562)."""

import importlib
from typing import Dict

# map attribute names to module paths under mordu.storeroom
_submodule_names: Dict[str, str] = {
    "fluids": "mordu.storeroom.fluids",
    "eos_purefluid": "mordu.storeroom.eos_purefluid",
    "cp0s": "mordu.storeroom.cp0s",
    "alpha_r_func": "mordu.storeroom.alpha_r_func",
}

# exported names visible from `from mordu.storeroom import ...`
__all__ = list(_submodule_names.keys())


def __getattr__(name: str):
    """Lazy import a submodule when accessed (e.g. storeroom.fluids)."""
    if name in _submodule_names:
        module = importlib.import_module(_submodule_names[name])
        # cache on the package module to avoid repeated imports
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(list(globals().keys()) + __all__)





import importlib
from mordu.purefluid import PureFluid

_lazy_objects = {
    "H2": lambda: PureFluid(
        name = "hydrogen",
        formula = "H2",
        M = 2.01588*1e-3,                 # [kg/mol], source = https://webbook.nist.gov/cgi/cbook.cgi?ID=C1333740
        P_c = 1.2964e6,                   # [Pa], source = [0313]
        T_c = 33.145,                     # [K], source = [0313]
        rho_c = 15.508*1e3,               # [mol/m3], source = [0313]
        P_t = 0.00736e6,                  # [Pa], source = [0313]
        T_t = 13.957,                     # [K], source = [0313]
        rho_t = 38.2*1e3,                 # [mol/m3], source = [0313]
        omega = -0.22,                    # [-], source = https://en.wikipedia.org/wiki/Acentric_factor
        sigma = 2.827*1e-10,              # [m], source = [0421] page  779
        epsilon = 59.7,                   # [1/K], source = [0421] page  779
        Sutherland_S = 97,                # [K], source = "Viscous Fluid Flow" page 28
        Sutherland_T0 = 273,              # [K], source = "Viscous Fluid Flow" page 28
        Sutherland_mu0 = 8.411e-6,        # [N s/m2], source = "Viscous Fluid Flow" page 28
        dipole = 0                        #it is an apolar molecule
        ),

    "N2": lambda: PureFluid(
        name = "nitrogen",
        formula = "N2",
        M = 28.0134e-3,                   # [kg/mol], source = https://webbook.nist.gov/cgi/cbook.cgi?ID=C7727379&Mask=4
        P_c = 33.978e5,                   # [Pa], source = [0532] page 65
        T_c = 126.192,                    # [K], source = [0451] page 44
        rho_c=  11.1839*1e3,              # [mol/m3], source = [0451] page 44
        P_t =  0.012563e6,                # [Pa], source = [0532] page 62
        T_t = 63.15,                      # [K], source = [0532] page 62
        rho_t = 31.046e3,                 # [mol/m3], source = [0532] page 62
        omega = 0.040,                    # [-], source = https://en.wikipedia.org/wiki/Acentric_factor
        sigma = 3.798e-10,                # [m], source = [0421] page 780
        epsilon = 71.4,                   # [1/K], source = [0421] page 780
        Sutherland_S = 107,               # [K], source = "Viscous Fluid Flow" page 28
        Sutherland_T0 = 273,              # [K], source = "Viscous Fluid Flow" page 28
        Sutherland_mu0 = 1.663e-5,        # [N s/m2], source = "Viscous Fluid Flow" page 28
        dipole = 0                        #it is an apolar molecule
        ),

    "NH3": lambda: PureFluid(
        name = "ammonia",
        formula = "NH3",
        M = 17.03052*1e-3,               # [kg/mol], source = https://webbook.nist.gov/cgi/cbook.cgi?ID=C7664417
        P_c = 11.3634e6,                 # [Pa], source = [0300] table 1
        T_c = 405.56,                    # [K], source = [0300] table 1
        rho_c =  13.696e3,               # [mol/m3], source = [0300] table 1
        P_t = 6.05339e3,                 # [Pa], source = [0300] table 1
        T_t = 195.49,                    # [K], source = [0300] table 1
        rho_t = 0.0037408e3,             # [mol/m3], source = [0300] table 1 (vapor density at triple point)
        omega = 0.253,                   # [-], source = https://en.wikipedia.org/wiki/Acentric_factor
        sigma = 2.9e-10,                 # [m], source = [0421] page 780
        epsilon = 558.3,                 # [1/K], source = [0421] page 780
        Sutherland_S = 377,              # [K], source = "Viscous Fluid FLow" page 577
        Sutherland_T0 = 273,             # [K], source = "Viscous Fluid FLow" page 577
        Sutherland_mu0 = 0.96e-5,        # [N s/m2], source = "Viscous Fluid FLow" page 577
        dipole = 1.47                    # [debye], source = [0421]  page 490
        ),

    "CH4": lambda: PureFluid(
        name = "methane",
        formula = "CH4",
        M = 16.04246e-3,                # [kg/mol], source = [0451]
        P_c = 4.5922e6,                 # [Pa], source = [0543]
        T_c = 190.564,                  # [K], source = [0543]
        rho_c = 10.139342719e3,         # [mol/m3], source = [0451]
        P_t = 0.011696e6,               # [Pa], source = [0543]
        T_t = 90.6941,                  # [K], source = [0543]
        rho_t = None,                   # [mol/m3], source = None
        omega = 0.0110,                 # [-], source = https://www.chemeo.com/cid/27-471-9/Methane
        sigma = 3.758e-10,              # [m], source = [0421] page 779
        epsilon = 148.6,                # [1/K], source = [0421] page 779
        Sutherland_S = 198,             # [K], source = "Viscous Fluid FLow" page 577
        Sutherland_T0 = 273,            # [K], source = "Viscous Fluid FLow" page 577
        Sutherland_mu0 = 1.1996e-5,     # [N s/m2], source = "Viscous Fluid FLow" page 577
        dipole = 0                      # it is an apolar molecule
        ),

    "C2H6": lambda: PureFluid(
        name = "ethane",
        formula = "C2H6",
        M =30.0690*1e-3 ,
        P_c = 49e5,
        T_c = 305.3,
        rho_c = 6.9e3,
        P_t = 0.000011e5,
        T_t = 91,
        rho_t = None,
        omega = 0.099,
        sigma = 4.443e-10,
        epsilon = 215.7,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 0
        ),

    "C8H18": lambda: PureFluid(
        name = "isooctane",
        formula = "C8H18",
        M = 114.2285e-3,
        P_c = 25.7e5,
        T_c = 543.8,
        rho_c =  2.14e3,
        P_t = 1,
        T_t = 165.76,
        rho_t = None,
        omega = 0.304,
        sigma = None,
        epsilon = None,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 0
        ),

    "CH3OH": lambda: PureFluid(
        name = "methanol",
        formula = "CH3OH",
        M = 32.0419e-3,
        P_c = 81e5,
        T_c = 513,
        rho_c = 8.51e3 ,
        P_t = 0.1835,
        T_t = 175.5,
        rho_t = 1.271e-7*1e3,
        omega = 0.565,
        sigma = 3.636e-10,
        epsilon = 481.8,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 1.7
        ),

    "C2H5OH": lambda: PureFluid(
        name = "ethanol",
        formula = "C2H5OH",
        M = 46.0684e-3,
        P_c = 63e5,
        T_c = 514,
        rho_c = 6e3,
        P_t = 7.185e-10*1e6,
        T_t = 150,
        rho_t = None,
        omega = 0.649,
        sigma = 4.53e-10,
        epsilon = 362.6,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 1.7
        ),

    "C4H9OH": lambda: PureFluid(
        name = "nbutanol",
        formula = "C4H8OH",
        M = 74.1216*1e-3,
        P_c = 44.23*1e5,
        T_c = 563.05,
        rho_c =  1/275*1e6,
        P_t = 1,
        T_t = 184.54,
        rho_t = None,
        omega = 0.59,
        sigma = None,
        epsilon = None,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 1.8
        ),

    "C7H16": lambda: PureFluid(
        name = "heptane",
        formula = "C7H16",
        M = 100.2019,
        P_c = 27.4*1e5,
        T_c = 540,
        rho_c =  2.35*1e3,
        P_t = 1,
        T_t = 182.56,
        rho_t = None,
        omega = 0.35,
        sigma = None,
        epsilon = None,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 0
        ),

    "H2O": lambda: PureFluid(
        name = "water",
        formula = "H2O",
        M = 18.0153e-3,
        P_c = 220.64e5,
        T_c = 647,
        rho_c = 17.9e3,
        P_t = 0.0061e5,
        T_t = 273.16,
        rho_t = None,
        omega = 0.344,
        sigma = 2.641e-10,
        epsilon = 809.1,
        Sutherland_S = 1064,
        Sutherland_T0 = 350,
        Sutherland_mu0 = 1.12e-5,
        dipole = 1.8
        ),

    "CO2": lambda: PureFluid(
        name = "carbon-dioxide",
        formula = "CO2",
        M = 44.0095e-3,
        P_c = 73.8,
        T_c = 304.18,
        rho_c = 10.6e3,
        P_t = 5.185e5,
        T_t = 216.58,
        rho_t = None,
        omega = 0.225,
        sigma = 3.941e-10,
        epsilon = 195.2,
        Sutherland_S = None,
        Sutherland_T0 = None,
        Sutherland_mu0 = None,
        dipole = 0
        ),
}

_loaded_objects = {}

def __getattr__(name: str):
    if name in _loaded_objects:
        return _loaded_objects[name]
    if name in _lazy_objects:
        obj = _lazy_objects[name]()
        _loaded_objects[name] = obj
        return obj
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

def __dir__():
    return sorted(list(globals().keys()) + list(_lazy_objects.keys()))



#03/03/2025
from .symbols import *

# reference state class, stored here because its only used here
class ReferenceState():

    def __init__(self, fluid, cp0_object, T0, T00, P0, P00):
        cp0_int_T = cp0_object.cp0_int_T
        cp0_over_T_int_T = cp0_object.cp0_over_T_int_T

        u_00 = 0
        h_00 = u_00 + fluid.R*T00 
        h0 = h_00 + cp0_int_T.subs([(T, T0)]) - cp0_int_T.subs([(T, T00)])

        #s_0 = s_00 + int(cp0_over_T(T_0, T_00)) - Rln(T_0/T_00) - Rln(rho_0/rho_00)
        #s_0 = s_00 + int(cp0_over_T(T_0, T_00)) - Rln(P_0/P_00)
        s_00 = 0
        s0 = s_00 + cp0_over_T_int_T.subs([(T, T0)]) - cp0_over_T_int_T.subs([(T, T00)]) - fluid.R*sp.log(P0/P00)

        self.P0 = P0
        self.T0 = T0
        self.h0 = h0
        self.s0 = s0

# ideal non dimensional helmholtz energy class
# every pure fluid EOS requires an object of this class for build
class Alpha0:

    def __init__(self, fluid, cp0_object, reference = None, alpha0 = None):
        #check if the fluid used in cp0 corresponds to the fluid
        if fluid.formula != cp0_object.fluid_formula:
            raise AttributeError("The fluid used in the cp0 object does not match the desired fluid")
        #handling cp0
        cp0_int_T = cp0_object.cp0_int_T
        cp0_over_T_int_T = cp0_object.cp0_over_T_int_T

        #handling reference state
        if reference == None:
            T0 = 300
            T00 = fluid.T_t
            P0 = 1e3
            P00 = fluid.P_t

        else:
            T0, T00, P0, P00 = reference

        rf = ReferenceState(fluid, cp0_object, T0, T00, P0, P00)

        #handling alpha0
        if alpha0 == None:
            alpha0 = 1/(fluid.R*T)*cp0_int_T + rf.h0/(fluid.R*T) -1 - 1/fluid.R*cp0_over_T_int_T + sp.ln(rho*T/(rf.P0/(fluid.R*rf.T0)*rf.T0)) - rf.s0/fluid.R

        self.alpha_0 = alpha0
        self.cp0 = cp0_object
        

